#include "Input.h"
#include <cstdlib>
#include <fstream>
#include <cassert>
#include <string>
#include <locale>
#include <sstream>
#include "csvgetlinec++.h"

template <typename T>
bool vector_contains(vector<T> vec, T element)
{
      return (std::find(vec.begin(), vec.end(), element) != vec.end());
}

Input::Input(string dir_name)
{
      // Insert the code that reads the input from the file
      // and stores it into the data structures of the object

      read_crs_table(dir_name);
      read_takes(dir_name);
      read_days(dir_name);
      read_preassigned(dir_name);
      read_paired(dir_name);

      crs_req.clear();
      crs2int.clear();
      int2crs.clear();
      int i = 0;
      for (pair<string, vector<int>> s : crs_day)
      { //.begin(), crs_day.end(), [this] (string s) {
            if (ALL_CRS || s.first[0] == 'D' || s.first[0] == 'M')
            {
                  crs_req[s.first] = s.second.size();
                  crs2int[s.first] = i++;
                  int2crs.push_back(s.first);
            }
      }
      for (pair<string, int> a : crs_req)
      {
            for (pair<string, int> b : crs_req)
            {
                  if (a.first != b.first)
                  {
                        int nstds = 0;
                        for (pair<string, vector<string>> s : stds)
                        {
                              if (find(s.second.begin(), s.second.end(), a.first) != s.second.end() and find(s.second.begin(), s.second.end(), b.first) != s.second.end())
                                    nstds++;
                        }
                        shared_stds.insert(make_pair(make_pair(a.first, b.first), nstds));
                  }
            }
      }
}

void Input::read_crs_table(string dir_name)
{
      string file_name = dir_name + "course_table.csv";
      ifstream in(file_name.c_str());
      if (!in)
      {
            cerr << "Cannot open file " << file_name << endl;
            exit(1);
      }
      string line;
      istringstream iss;

      cout << "* Reading " << file_name << "... " << flush;

      int linenum = 0;
      Csv csv(in, "\t");

      crs_table.clear();
      while (csv.getline(line) != 0)
      {
            linenum++;
            if (csv.getnfield() < 2 || (csv.getfield(0) == "") || (csv.getfield(0) == "#"))
            {
                  cerr << "Skip line " << linenum << endl;
                  continue;
            }
            string crs_id = csv.getfield(0);
            string crs_str = csv.getfield(1);
            size_t next = crs_str.find_first_of(",", 0);
            string crs_code = crs_str.substr(0, next);
            crs_code.erase(crs_code.find_last_not_of(" \n\r\t") + 1);

            crs_table.insert(make_pair(crs_id, crs_code));
            //crs_table[crs_id] = crs_code;
      }
      cout << linenum << " lines read" << endl;
}

void Input::read_takes(string dir_name)
{

      string file_name = dir_name + "takes.csv";
      ifstream in(file_name.c_str());
      if (!in)
      {
            cerr << "Cannot open file " << file_name << endl;
            exit(1);
      }
      string line;
      istringstream iss;

      cout << "* Reading " << file_name << "... " << flush;

      int linenum = 0;
      Csv csv(in, "\t");

      stds.clear();
      while (csv.getline(line) != 0)
      {
            linenum++;
            if (csv.getnfield() < 2 || (csv.getfield(0) == "") || (csv.getfield(0) == "#"))
            {
                  cerr << "Skip line " << linenum << endl;
                  continue;
            }
            string std_id = csv.getfield(0);
            string crs_id = csv.getfield(1);
            std_id.erase(std_id.find_last_not_of(" \n\r\t") + 1);
            crs_id.erase(crs_id.find_last_not_of(" \n\r\t") + 1);
            //cout << "+" << crs_id << "_" << std_id << "+" << endl;
            if (stds.find(std_id) == stds.end())
            {
                  //vector<string> tmp={std_id};
                  //stds.insert(pair<string, vector<string> > (crs_id,{}));
                  //stds[crs_id].clear();
                  stds[std_id].push_back(crs_table[crs_id]);
            }
            else
            {
                  stds[std_id].push_back(crs_table[crs_id]);
            }
      }
      cout << linenum << " lines read" << endl;
}

void Input::read_days(string dir_name)
{

      string file_name = dir_name + "days.csv";
      ifstream in(file_name.c_str());
      if (!in)
      {
            cerr << "Cannot open file " << file_name << endl;
            exit(1);
      }
      string line;
      istringstream iss;

      cout << "* Reading " << file_name << "... " << flush;

      int linenum = 0;
      Csv csv(in, ",");

      days.clear();
      while (csv.getline(line) != 0)
      {
            linenum++;
            if ((csv.getfield(0) == "") || (csv.getfield(0) == "#"))
            {
                  cerr << "Skip line " << linenum << endl;
                  continue;
            }
            for (int i = 0; i < csv.getnfield(); i++)
            {
                  string day = csv.getfield(i);
                  bool has_only_digits = (day.find_first_not_of("0123456789") == string::npos);
                  if (has_only_digits)
                  {
                        int n = atoi(day.c_str());
                        if (n <= 31 && n >= 1)
                              days.push_back(n);
                        else
                        {
                              cerr << day << " not a number from 1 to 31\n";
                              exit(0);
                        }
                  }
                  else
                  {
                        cerr << day << " not a number from 1 to 31\n";
                        exit(0);
                  }
            }
      }
      cout << linenum << " lines read" << endl;
}

void Input::read_preassigned(string dir_name)
{

      string file_name = dir_name + "dates.txt";
      ifstream in(file_name.c_str());
      if (!in)
      {
            cerr << "Cannot open file " << file_name << endl;
            exit(1);
      }
      string line;
      istringstream iss;

      cout << "* Reading " << file_name << "... " << flush;

      int linenum = 0;
      crs_day.clear();
      int day_num = -1;
      while (!in.eof())
      {
            linenum++;
            getline(in, line);
            //cout<<line<<endl;
            if (line.length() == 0 || line[0] == '#')
                  continue;
            std::locale loc;
            if (isdigit(line[0], loc))
            {
                  size_t next = line.find_first_of(".", 0);
                  string day = line.substr(0, next);
                  day_num = atoi(day.c_str());
            }
            else
            {
                  string crs_code = line.erase(line.find_last_not_of(" \n\r\t") + 1);
                  if (crs_code[crs_code.size() - 1] == 'm')
                  {
                        crs_code = crs_code.erase(crs_code.size() - 1);
                  }
                  crs_day[crs_code].push_back(day_num);
            }
      }
      cout << linenum << " lines read" << endl;
}

void Input::read_paired(string dir_name)
{
      string file_name = dir_name + "paired_courses.csv";
      ifstream in(file_name.c_str());
      if (!in)
      {
            cerr << "Cannot open file " << file_name << endl;
            exit(1);
      }
      string line;
      istringstream iss;

      cout << "* Reading " << file_name << "... " << flush;

      int linenum = 0;
      Csv csv(in, ",");

      paired.clear();
      while (csv.getline(line) != 0)
      {
            linenum++;
            if ((csv.getfield(0) == "") || (csv.getfield(0) == "#"))
            {
                  cerr << "Skip line " << linenum << endl;
                  continue;
            }
            set<string> tmp;
            for (int i = 0; i < csv.getnfield(); i++)
            {
                  string crs = csv.getfield(i);
                  if (crs != "")
                  {
                        tmp.insert(crs);
                  }
            }
            paired.push_back(tmp);
      }
      cout << linenum << " lines read" << endl;
}


ostream &operator<<(ostream &os, const Input &pa)
{

      os << "days = [";
      for_each(pa.days.begin(), pa.days.end(), [&](int s) {
            os << s << " ";
      });
      os << "]\n";

      os << "courses = {";
      for_each(pa.crs_req.begin(), pa.crs_req.end(), [&](pair<string, int> s) {
            os << s.first << " " << s.second << "; ";
      });
      os << "}\n";
      os << "paired = {";
      for_each(pa.paired.begin(), pa.paired.end(), [&](set<string> s) {
            os << "[ ";
            for_each(s.begin(), s.end(), [&](string crs) { os << crs << " "; });
            os << "] ";
      });
      os << "}\nshared_stds: ";
      int count = 0;
      for (pair<string, int> a : pa.crs_req)
      {
            for (pair<string, int> b : pa.crs_req)
            {
                  if (a.first != b.first)
                  {
                        count++;
                        os << "(" << a.first << "," << b.first << ") " << pa.getNShared(a.first, b.first) << " ";
                  }
            }
            if (count > 20)
            {
                  os << "...";
                  break;
            }
      }
      os << "\n# courses: " << pa.getNCourses() << " # days: " << pa.getNDays() << endl;

      return os;
}
