#ifndef INPUT_HH
#define INPUT_HH

#include <iostream>
#include <vector>
#include <unordered_map>
#include <map>
#include <set>
#include <cmath>

#define ALL_CRS 0

using namespace std;


class Input {
   friend ostream& operator<<(ostream& os, const Input& bs);
 public:
   Input(string file_name);

   const int getNCourses() const {
      return crs_req.size();
   }
   const int getNDays() const {
      return days.size();
   }
   const int getNShared(string a, string b) const {
      return shared_stds.at(make_pair(a,b));
   }

   const int getDistance(int d1, int d2) const {
      return abs(d1-d2);
   }
	
// protected:
   vector<int> days;
   unordered_map<string, vector<int> > crs_day;
   unordered_map<string, int> crs_req;
   unordered_map<string, int> crs2int;
   vector<string> int2crs;
   unordered_map<int, set<string> > cliques;
   vector< set<string> > paired;

 private:
   void read_crs_table(string dir_name);
   void read_takes(string dir_name);
   void read_days(string dir_name);
   void read_preassigned(string dir_name);
   void read_paired(string dir_name);

   unordered_map<string, string> crs_table;
   unordered_map<string, vector<string> > stds;
   map< pair<string, string>, int> shared_stds;
};

#endif
